using MeshVR;
using MVR.FileManagementSecure;
using System.Collections;
using System.Collections.Generic;
using UnityEngine;



public class SkinMicroDetail : MVRScript
{
    private DAZCharacterSelector selector;
    private DAZCharacter dazCharacter;


    private const string DEFAULT_DETAIL_MAP = "Custom/Atom/Person/Textures/Skin81.png";


    //  Excluded slots 12 and 14 for finger and toenails
    private int[] femaleSkinSlots = new int[] { 15, 18, 19, 20, 21, 27, 29, 2, 5, 11, 0, 16, 17, 22, 23, 28 };
    //  Excluded slots 12 and 14 for finger and toenails
    private int[] maleSkinSlots = new int[] { 15, 16, 19, 20, 21, 22, 30, 32, 2, 5, 11, 0, 17, 18, 23, 24, 28, 29 };

    private bool isLoaded = false;

    private Dictionary<int, Shader> originalShaders = new Dictionary<int, Shader>();
    private Texture2D detailMap;

    private JSONStorableFloat tesselation = new JSONStorableFloat("Tesselation", 3.75f, 1.0f, 8.0f);
    private JSONStorableFloat phongStrength = new JSONStorableFloat("Phong Strength", 0.5f, 0.0f, 1.0f);
    private JSONStorableUrl detailMapURL = new JSONStorableUrl("Detail Map Texture", string.Empty);

    private JSONStorableFloat detailWeight = new JSONStorableFloat("Detail Weight", 0.25f, 0.0f, 1.0f);
    private JSONStorableFloat detailMapScale = new JSONStorableFloat("Detail Scale", 50.0f, 0.0f, 200.0f, false);

    public override void Init()
    {

        if (containingAtom == null || containingAtom.type != "Person")
        {
            SuperController.LogError("This only works with person Atoms");
            DestroyImmediate(this);
            return;
        }

        selector = containingAtom.GetComponentInChildren<DAZCharacterSelector>();
        dazCharacter = selector.selectedCharacter;

        RegisterFloat(tesselation);
        tesselation.setCallbackFunction += (float val) => UpdateShaderParams();
        CreateSlider(tesselation);
        RegisterFloat(phongStrength);
        phongStrength.setCallbackFunction += (float val) => UpdateShaderParams();
        CreateSlider(phongStrength);

        RegisterUrl(detailMapURL);
        detailMapURL.setCallbackFunction += LoadDetailMap;


        UIDynamicButton buttonSecondTex = CreateButton("Browse Detail Map", false);
        buttonSecondTex.button.onClick.AddListener(() =>
        {
            // Todo: Most recent DIR, or default
            List<ShortCut> shortCuts = FileManagerSecure.GetShortCutsForDirectory(GetPackagePath(this) + "Custom/Atom/Person/Textures", false, true, true, false);
            SuperController.singleton.GetMediaPathDialog((string path) => { detailMapURL.val = path; }, "jpg|png|tif|tiff", GetPackagePath(this) + "Custom/Atom/Person/Textures", false, true, false, null, true, shortCuts, true, true);

        });
        CreateTextField(detailMapURL, false);


        RegisterFloat(detailWeight);
        detailWeight.setCallbackFunction += (float val) => UpdateShaderParams();
        CreateSlider(detailWeight);

        RegisterFloat(detailMapScale);
        detailMapScale.setCallbackFunction += (float val) => UpdateShaderParams();
        CreateSlider(detailMapScale);


        if (detailMapURL.val.Equals(string.Empty))
        {

            detailMapURL.SetFilePath(GetPackagePath(this) + DEFAULT_DETAIL_MAP);

        }

        StartCoroutine(Setup());
    }

    private void LoadDetailMap(string path)
    {
        ImageLoaderThreaded.QueuedImage queuedImage = new ImageLoaderThreaded.QueuedImage();
        queuedImage.imgPath = path;
        queuedImage.forceReload = true;
        queuedImage.skipCache = true;
        queuedImage.isNormalMap = true;
        queuedImage.linear = true;
        queuedImage.compress = false;
        queuedImage.createMipMaps = false;
        queuedImage.callback = (ImageLoaderThreaded.QueuedImage qi) =>
        {
            detailMap = qi.tex;
            UpdateShaderParams();
        };
        ImageLoaderThreaded.singleton.QueueImage(queuedImage);
    }

    private IEnumerator Setup()
    {
        while (!dazCharacter || !dazCharacter.ready)
        {
            yield return null; // Wait for a frame
        }

        string shaderName = Shader.Find("Custom/Subsurface/GlossNMDetailTessMappedComputeBuff").name;
        foreach (int i in (GetMaterialSlots()))
        {
            if (!shaderName.Equals(dazCharacter.skin.GPUmaterials[i].shader.name))
            {
                originalShaders[i] = dazCharacter.skin.GPUmaterials[i].shader;
            }
        }

        foreach (int i in (GetMaterialSlots()))
        {
            Material mat = dazCharacter.skin.GPUmaterials[i];
            mat.shader = Shader.Find("Custom/Subsurface/GlossNMDetailTessMappedComputeBuff");
        }
        UpdateShaderParams();
        dazCharacter.skin.BroadcastMessage("OnApplicationFocus", true);
        isLoaded = true;
    }

    public void Update()
    {
        if (selector && selector.selectedCharacter != dazCharacter)
        {
            dazCharacter = selector.selectedCharacter;
            StartCoroutine(Setup());
            return;
        }

        // Detect Shader Reset on scene load, appearance preset load, etc.
        if (dazCharacter && dazCharacter.ready && (!originalShaders.ContainsKey(0) || !originalShaders[0].name.Equals(dazCharacter.skin.GPUmaterials[0].shader.name)))
        {
            StartCoroutine(Setup());
        }
    }
    private void UpdateShaderParams()
    {
        foreach (int i in (GetMaterialSlots()))
        {
            Material mat = dazCharacter.skin.GPUmaterials[i];
            mat.SetTexture("_TessTex", Texture2D.whiteTexture);
            mat.SetFloat("_TessPhong", phongStrength.val);
            mat.SetFloat("_Tess", tesselation.val);
            mat.SetTexture("_DetailMap", detailMap);
            mat.SetFloat("_DetailWeight", detailWeight.val);
            mat.SetTextureScale("_DetailMap", Vector2.one * detailMapScale.val);
        }
    }

    private int[] GetMaterialSlots()
    {
        return dazCharacter.isMale ? maleSkinSlots : femaleSkinSlots;
    }

    public void OnDisable()
    {
        if (isLoaded && dazCharacter.ready)
        {
            foreach (int slot in GetMaterialSlots())
            {
                dazCharacter.skin.GPUmaterials[slot].shader = originalShaders[slot];
            }
            dazCharacter.skin.BroadcastMessage("OnApplicationFocus", true);
        }
    }

    public void OnEnable()
    {
        StartCoroutine(Setup());
    }

    public void OnDestroy()
    {
        OnDisable();
        StopAllCoroutines();
    }

    // Get directory path where the plugin is located. Based on Alazi's & VAMDeluxe's method.
    // Credit to MacGruber Utils shared under CC BY
    private static string GetPluginPath(MVRScript self)
    {
        string id = self.name.Substring(0, self.name.IndexOf('_'));
        string filename = self.manager.GetJSON()["plugins"][id].Value;
        return filename.Substring(0, filename.LastIndexOfAny(new char[] { '/', '\\' }));
    }

    // Get path prefix of the package that contains our plugin.
    // Credit to MacGruber Utils shared under CC BY
    private static string GetPackagePath(MVRScript self)
    {
        string filename = GetPluginPath(self);
        int idx = filename.IndexOf(":/");
        if (idx >= 0)
            return filename.Substring(0, idx + 2);
        else
            return string.Empty;
    }


}